/*
 * Vifir - An electronic flight bag wannabe utility for aircraft pilots
 * Copyright (C) 2009 David Ludlow
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * dave@adsllc.com, PO Box 13070, Toledo, OH 43623
 */

#ifdef HAVE_CONFIG_H
 #include "../config.h"
#endif

#include <math.h>
#ifndef WIN32
#include <pthread.h>
#endif
#include <iostream>
#include "GPS.h"
#include "Util.h"

/* This stuff needs to be global so it's visible within the class
 * and within the callback
 */
volatile int GPS_Mutex=0;
/* In theory, the mutex should prevent these from being volatile, but it's such
 * a tiny penalty we'll just be extra safe
 */
volatile double Lon,Lat,Alt,Trk,Spd;
volatile time_t LonLatTime;		/* Time Lon & Lat were received */
volatile time_t AltTime;		/* Time altitude was received */
volatile time_t VelTime;		/* Time velocity (Trk/Spd) was received */

GPS::GPS(Config *Cfg)
{
#ifdef HAVE_LIBGPS
	gpsddata=NULL;
#endif
	ActiveFlag=false;
	this->Cfg=Cfg;
}

GPS::~GPS()
{
}

bool GPS::Activate() {
#ifdef HAVE_LIBGPS
	switch (Cfg->GetGPSSource())
	{
		case Config::gpsd:
			/* http://gpsd.mainframe.cx/ and http://gpsd.rellim.com/ have gpsd data sources we can connect to for testing */
			gpsddata=gps_open("127.0.0.1","2947");
			if (!gpsddata) {
				return false;
			}
			gps_set_callback(gpsddata,GPS_Callback,&handler);
			break;
		case Config::Simulated:
			Lon=Cfg->GetGPSSimLon();
			Lat=Cfg->GetGPSSimLat();
			Spd=KT2MPERS(Cfg->GetGPSSimSpd());
			Trk=Cfg->GetGPSSimTrk();
			Alt=FT2METERS(Cfg->GetGPSSimAlt());
			LonLatTime=time(NULL);
			VelTime=LonLatTime;
			AltTime=LonLatTime;
			break;
		case Config::Serial:
			break;
	}

	ActiveFlag=true;
#endif
	return true;
}

void GPS::Deactivate() {
#ifdef HAVE_LIBGPS
	if (gpsddata) {
		gps_del_callback(gpsddata,&handler);
		gps_close(gpsddata);
		gpsddata=NULL;
	}
	ActiveFlag=false;
#endif
}

bool GPS::GetFix(double *Lon, double *Lat, double *Alt, double *Trk, double *Spd) {
	GPS_Mutex++;
	if (!CheckMutex()) {
		GPS_Mutex--;
		return false;
	}

	if (Cfg->GetGPSSource()==Config::Simulated)
	{
		time_t deltat;
		deltat=time(NULL)-LonLatTime;

		::Spd=KT2MPERS(Cfg->GetGPSSimSpd());
		::Trk=Cfg->GetGPSSimTrk();
		::Alt=FT2METERS(Cfg->GetGPSSimAlt());

		/* This is a simple linear approximation.  A proper job would handle the
		 * coordinate system projection.  For that matter, it would do something
		 * with great circles too.  Since this is just a simulation, we don't
		 * bother.
		 */
		::Lon+=(sin(::Trk*(M_PI/180.0))*::Spd*(deltat/3600.0))/60.0;
		::Lat+=(cos(::Trk*(M_PI/180.0))*::Spd*(deltat/3600.0))/60.0;

		LonLatTime=time(NULL);
		VelTime=LonLatTime;
		AltTime=LonLatTime;
	}

	/* We don't want any data more than 2 seconds old */
	time_t tooold;
	tooold=time(NULL)-2;

	if (Alt && AltTime<tooold) {
		GPS_Mutex--;
		return false;
	}
	if ((Lon || Lat) && LonLatTime<tooold) {
		GPS_Mutex--;
		return false;
	}
	if ((Trk || Spd) && VelTime<tooold) {
		GPS_Mutex--;
		return false;
	}

	if (Lon)
		*Lon=::Lon;
	if (Lat)
		*Lat=::Lat;
	if (Alt)
		*Alt=::Alt;
	if (Trk)
		*Trk=::Trk;
	if (Spd)
		*Spd=::Spd;
	GPS_Mutex--;
	return true;
}

bool GPS::GetActive() {
	return ActiveFlag;
}

bool GPS::CheckMutex() {
	for (int i=0;i<5;i++) {
		if (GPS_Mutex==1) {
			return true;
		}
		if (i!=4) {	/* No sense sleeping after the last check */
#ifndef WIN32
			usleep(10);
#endif
		}
	}
	return false;
}

#ifdef HAVE_LIBGPS
void GPS_Callback(gps_data_t *sentence, char *buf, size_t size, int i) {
	GPS_Mutex++;
	if (GPS_Mutex==1) {
		/* Store the data for later use */
		time_t now=time(NULL);
		if (!isnan(sentence->fix.longitude) && !isnan(sentence->fix.latitude)) {
			LonLatTime=now;
			Lon=sentence->fix.longitude;
			Lat=sentence->fix.latitude;
		}
		if (!isnan(sentence->fix.altitude)) {
			AltTime=now;
			Alt=sentence->fix.altitude;
		}
		if (!isnan(sentence->fix.track) && !isnan(sentence->fix.speed)) {
			VelTime=now;
			Trk=sentence->fix.track;
			Spd=sentence->fix.speed;
		}
	}
	GPS_Mutex--;
}
#endif
